#include <sys/stat.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include <octave/oct.h>
#include <octave/dColVector.h>
#include <octave/dMatrix.h>
#include <octave/EIG.h>
#include "channelflow/flowfield.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"

#include "channelflow/quicksort.h"
#include "octutils.h"

using namespace std;
using namespace channelflow;

// This program calculates eigenvalues of fixed point of plane Couette flow
// using Arnoldi iteration. The ideas and algorithm are based on Divakar
// Viswanath, "Recurrent motions within plane Couette turbulence",
// J. Fluid Mech.</em> 580 (2007), http://arxiv.org/abs/physics/0604062.
// Notation and details of the algorithm are based on Lecture 33 of
// "Numerical Linear Algebra" by Trefethen and Bau. Production runs began
// in early December 2006.

// Arnoldi iteration estimates the eigenvalues of a matrix A by iteratively
// constructing a QR decomposition of a matrix whose columns are
// Ab, A^2 b, etc., where b is an arbitrary starting vector.

// For the present problem, A is a discretized form of PCF dynamics,
// linearized about the equilibrium solutions, and b is the discretized
// form of an arbitrary starting velocity field.

// In continuous terms, the matrix A corresponds to a differential operator
// L and b to a perturbation du from an invariant solution u* of Navier-Stokes. 
//
// Let f^T be the time-T Navier-Stokes map 
//   f^T : u(t) -> u(t+T), 
//
// Let u*,T,sigma be an equilibrium solution of f (e.g. a relative periodic orbit)
//   u* = sigma f^T(u*). 
//
// Define the (approximately) linear operator L as
//  L du = sigma f^T(u* + du) - sigma f^T(u*) =  sigma f^T(u* + du) - u*
//
// This program computes L du, L^2 du , L^3 du, etc. with a DNS algorithm
// and translates between velocity fields du^n = L^n du and vectors
// v^n = A^n b via by choosing a set of linearly independent spectral 
// coefficients from the expansion for du

// f^T(u)  = time-T DNS integration of u,

// sf^T(u) = sigma f^T(u), where sigma is a symmetry of the flow


void sf(const FlowField& u, Real T, const FieldSymmetry& sigma, FlowField& sfu, 
	Real Reynolds, DNSFlags& flags, TimeStep& dt, int& fcount);

void Dsf(const FlowField& u, const FlowField& sfu, const FlowField& du, FlowField& Df_du,
	 Real T, const FieldSymmetry& sigma, Real Reynolds, DNSFlags& flags,
	 TimeStep& dt, Real eps, bool centerdiff, int& fcount);

void checkConjugacy(const ComplexColumnVector& u, const ComplexColumnVector& v);

int main(int argc, char* argv[]) {

  string purpose("computes eigenvalues of equilibrium or orbit with Arnoldi iteration");
  ArgList args(argc, argv, purpose);

  DNSFlags dnsflags;

  const bool poincare = args.getflag("-poinc",  "--poincare",    "computing eigenvalues of map on I-D=0 Poincare section");
  const int  Narnoldi = args.getint ("-N",  "--Narnoldi", 100, "number of Arnoldi iterations");
  const int  Nsave    = args.getint ("-Ne", "--Neigval", 20, "number of eigenvalues/functions to save");

  const Real T        = args.getreal("-T",  "--maptime", 10.0, "integration time for Navier-Stokes.");
  const string sigstr = args.getstr ("-sigma", "--sigma", "", "file containing sigma of sigma f^T(u) - u = 0 (default == identity)");
  const string symmstr = args.getstr ("-symms", "--symmetries", "", "file containing generators of isotropy group for symmetry-constrained search");
  const Real Reynolds = args.getreal("-R", "--Reynolds", 400, "Reynolds number");

  const Real EPS_arn  = args.getreal("-ek", "--epsKrylov", 1e-10, "min. condition # of Krylov vectors");
  const Real EPS_du   = args.getreal("-edu","--epsdu", 1e-7, "magnitude of Arnoldi perturbation");
  const bool centdiff = args.getflag("-cd", "--centerdiff", "centered differencing to estimate differentials");
  const int  seed     = args.getint ("-sd",  "--seed",       1, "seed for random number generator");
  const Real smooth   = args.getreal("-s",  "--smoothness", 0.4, "smoothness of initial perturb, 0 < s < 1");

  dnsflags.baseflow     = PlaneCouette;
  dnsflags.timestepping = s2stepmethod(args.getstr("-ts", "--timestepping", "SBDF3", "time step algorithm: sbdf[1234]|cnab2|cnrk2|smrk2"));
  dnsflags.dealiasing   = DealiasXZ;
  dnsflags.nonlinearity = s2nonlmethod(args.getstr("-nl", "--nonlinearity", "Rotational", "nonlinearity method: rot|conv|skew|div|alt"));
  dnsflags.constraint   = PressureGradient;
  dnsflags.dPdx         = 0.0;  
  dnsflags.verbosity    = Silent;

  const bool vardt     = args.getflag("-vdt","--variableDt", "adjust dt to keep CFLmin<=CFL<CFLmax");
  const Real dtarg     = args.getreal("-dt","--timestep",  0.03125, "(initial) integration timestep");
  const Real dtmin     = args.getreal("-dtmin", "--dtmin", 0.001, "minimum timestep");
  const Real dtmax     = args.getreal("-dtmax", "--dtmax", 0.04,  "maximum timestep");
  const Real dTCFL     = args.getreal("-dTCFL", "--dTCFL", 1.00,  "check CFL # at this interval");
  const Real CFLmin    = args.getreal("-CFLmin", "--CFLmin", 0.40, "minimum CFL number");
  const Real CFLmax    = args.getreal("-CFLmax", "--CFLmax", 0.60, "maximum CFL number");

  const string duname = args.getstr ("-du", "--perturb", "", "initial perturbation field, random if unset");
  const string outdir = args.getpath("-o", "--outdir", "./", "output directory");

  const string uname  = args.getstr(1, "<flowfield>", "the equilibrium field");

  args.check();
  args.save("./");
  mkdir(outdir);

  //if (int(eqb) + int(orb) != 1)
  //cferror("please choose either -orb or -eqb flag");

  srand48(seed);
  //const Real nu = 1.0/Reynolds;
  const Real decay = 1.0-smooth;
  const int prec = 16;

  PoincareCondition* h = poincare ? new DragDissipation() : 0;

  TimeStep dt(dtarg, dtmin, dtmax, dTCFL, CFLmin, CFLmax, vardt);

  if (!poincare)
    dt.adjust_for_T(T);

  FieldSymmetry sigma;       // defaults to identity
  if (sigstr.length() != 0)
    sigma = FieldSymmetry(sigstr);

  if (symmstr.length() > 0) {
    SymmetryList symms(symmstr);
    cout << "Restricting flow to invariant subspace generated by symmetries" << endl;
    cout << symms << endl;
    dnsflags.symmetries = symms;
  }


  int fcount = 0;

  FlowField u(uname);
  //project(dnsflags.symmetries, u, "initial field u", cout);
  //fixdivnoslip(u);

  const int Nx = u.Nx();
  const int Ny = u.Ny();
  const int Nz = u.Nz();

  const Real Lx = u.Lx();
  const Real ya = u.a();
  const Real yb = u.b();
  const Real Lz = u.Lz();

  const int kxmin = -u.kxmaxDealiased();
  const int kxmax =  u.kxmaxDealiased();
  const int kzmin =  0;
  const int kzmax =  u.kzmaxDealiased();

  cout << setprecision(17);
  cout << "   Nx == " << Nx << endl;
  cout << "   Ny == " << Ny << endl;
  cout << "   Nz == " << Nz << endl;
  cout << "kxmin == " << kxmin << endl;
  cout << "kxmax == " << kxmax << endl;
  cout << "kzmin == " << kzmin << endl;
  cout << "kzmax == " << kzmax << endl;
  //cout << "building basis (this takes a long time)..." << endl;

  // Used to communicate timestep parameters to integration functions.
  cout << "dt     == " << dtarg << endl;
  cout << "dtmin  == " << dtmin << endl;
  cout << "dtmax  == " << dtmax << endl;
  cout << "CFLmin == " << CFLmin << endl;
  cout << "CFLmax == " << CFLmax << endl;


  // Set up DNS operator ("A" in Arnoldi A*b terms)
  cout << "setting up DNS and initial fields..." << endl;
  ChebyTransform trans(Ny);

  Real CFL = 0.0;

  const Real eps = EPS_du/L2Norm(u);
  if (dnsflags.nonlinearity == LinearAboutField)
    u.setToZero();
  FlowField sfu(u);
  cout << "computing sigma f^T(u)..." << endl;
  sfp(u, T, Reynolds, h, sigma, sfu, dnsflags, dt, fcount, CFL, cout);
  cout << "\nCFL == " << CFL << endl;
  //sfu *= sigma;

  
  FlowField du(Nx, Ny, Nz, 3, Lx, Lz, ya, yb);
  FlowField u_du(Nx, Ny, Nz, 3, Lx, Lz, ya, yb);     // u + du
  FlowField sfu_du(Nx, Ny, Nz, 3, Lx, Lz, ya, yb);   // sigma f^T(u_du)
  FlowField Dsf_du(Nx, Ny, Nz, 3, Lx, Lz, ya, yb);   // sigma f^T(u*+du) - sigma f(u*)


  // Double-check that u,sigma,T are a solution of sigma f^T(u) = u
  du = u;
  du -= sfu;
  cout << "L2Norm(u - sigma f^T(u)) == " << L2Norm(du) << endl;

  if (L2Norm(du) > 1e-06) 
    cferror("error: u,sigma,T is not a solution of sigma f^T(u) - u = 0");

  du.setToZero();

  // Set du  = EPS_du (random unit perturbation, "b" in Arnoldi A*b terms)
  if (duname.length() == 0) {
    du.addPerturbations(kxmax, kzmax, 1.0, decay);
    du.addPerturbation(0, 0, 1.0, decay);
  }
  else
    du = FlowField(duname);

  du *= EPS_du/L2Norm(du);
  cout << "Post-rescaling..." << endl;


  vector<RealProfileNG> basis;

  // Project field du onto basis, producing vector b
  ColumnVector b; //(M);
  field2vector(du,b, basis);
  //field2vector(du,b);
  b *= 1.0/L2Norm(b);

  Arnoldi arnoldi(b, Narnoldi, EPS_arn);

  // Initialize output streams
  ofstream osLamconv((outdir + "Lamconv.asc").c_str());
  ofstream oslamconv;

  oslamconv.open((outdir + "lamconv.asc").c_str());
  oslamconv << setprecision(17);


  for (int n=0; n<Narnoldi; ++n) {
    cout << "======================================================" << endl;
    cout << n << "th iteration:" << endl;

    // Compute v = Aq in Arnoldi iteration terms, where q test vector
    // In fluid terms, this is
    //      L du == F^T(u + du) - F^T(u)
    // where |du| << |u|, so that RHS expression is approximately linear
    // and where f^T is the forward map for time T of Navier-Stokes

    // Relation of v and du :
    //   du == eps * v_m phi_m
    //    v == 1/eps (du, phi_m)

    const ColumnVector& q = arnoldi.testVector();

    cout << "building field du..." << endl;
    vector2field(q, du, basis);

    cout << "computing Df du..." << endl;
    // Compute Dsf du == 1/eps (sigma f(u + eps du) - sigma f(u)) or centerdiff equiv
    //Dsf(u, T, sigma, sfu, du, Dsf_du, Reynolds, dnsflags, dt, eps, centdiff, fcount, CFL, cout);
    Dsfp(u, T, Reynolds, h, sigma, sfu, du, Dsf_du, dnsflags, dt, eps, centdiff, fcount, CFL, cout);

    cout << "\nrexpressing Df du as column vector..." << endl;
    ColumnVector Aq; //(M);
    field2vector(Dsf_du, Aq, basis);

    // Send the product Aq back to Arnoldi
    cout << "doing LU decomp on Krylov subspace vectors..." << endl;
    arnoldi.iterate(Aq);

    cout << "results of LU decomp..." << endl;
    // Now do some output: Lambda, the eigenvalues of map f^T,
    ComplexColumnVector Lambda = arnoldi.ew();
    ComplexMatrix Vn = arnoldi.ev();
    int nsave=lesser(Nsave, n+1);

    cout << "\nabs, arg Lambda == " << endl;
    for (int j=0; j<nsave; ++j) {
      cout << setw(prec+5) << abs(Lambda(j)) << setw(prec+5) << arg(Lambda(j)) << endl;
      osLamconv << Re(Lambda(j)) << ' ' << Im(Lambda(j)) << ' ';
    }
    for (int j=nsave; j<Nsave; ++j)
      osLamconv << "-99 0 ";
    osLamconv << endl;

    ofstream osLam((outdir + "Lambda.asc").c_str());
    osLam << setprecision(17);
    for (int j=0; j<n; ++j)
      osLam     << Re(Lambda(j)) << ' ' << Im(Lambda(j)) << endl;
    osLam.close();

    ofstream osLam2((outdir + "LambdaAbsArg.asc").c_str());
    osLam2 << setprecision(17);
    for (int j=0; j<n; ++j)
      osLam2     << abs(Lambda(j)) << ' ' << arg(Lambda(j)) << endl;
    osLam2.close();

    // Convert eigvals of map F^T (Lambda) to eigvals of flow f (lambda)
    // Lambda == e^{lambda T}.
    // lambda == 1/T log Lambda
    //        == 1/T (log(abs(Lambda)) + i arg(Lambda))
    //

      ComplexColumnVector lambda(n+1);
      for (int j=0; j<=n; ++j)
	lambda(j) = (1.0/T) * log(Lambda(j));

      cout << "\nlambda == " << endl;
      for (int j=0; j<nsave; ++j) {
	cout      << Re(lambda(j)) << ' ' << Im(lambda(j)) << endl;
	oslamconv << Re(lambda(j)) << ' ' << Im(lambda(j)) << ' ';
      }
      // For file lamconv.asc, supplement computed eigvals with bogus values
      // -99 + 0 i so that all rows have same number of columns
      for (int j=nsave; j<Nsave; ++j)
	oslamconv << "-99 0 ";
      oslamconv << endl;

      // Save all estimated eigenvalues
      ofstream oslam((outdir + "lambda.asc").c_str());
      oslam << setprecision(17);
      for (int j=0; j<n; ++j)
      oslam     << Re(lambda(j)) << ' ' << Im(lambda(j)) << endl;
      oslam.close();



    // Print the period and char. mult of first oscillatory mdoe
    /*****************************************************
    for (int j=0; j<nsave; ++j) {
      if (abs(Im(lambda(j)) > 0.0)) {
	Real period  = abs(2*pi/Im(lambda(j)));
	Real magnify = exp(Re(lambda(j)*abs(period)));
	cout << "period  = " << period << endl;
	cout << "magnify = " << magnify << endl;
	break;
      }
    }
    *************************************************/

    // Reconstruct and save top ten eigenfunctions
    for (int j=0; j<nsave; ++j) {

      string sj1 = i2s(j+1);
      string sj2 = i2s(j+2);

      // Real-valued eigenvalue
      if (Im(Lambda(j)) == 0.0) {
	FlowField ef(u); // set ef to have same size as u
	ColumnVector ev = real(Vn.column(j));
	//vector2field(ev, ef);
	vector2field(ev, ef, basis);
	ef *= 1.0/L2Norm(ef);
	ef.save(outdir + "ef" + sj1);
	save(Lambda(j), outdir + "eW" + sj1);
	  Complex lambda = log(Lambda(j)) / T;
	  save(lambda, outdir + "ew" + sj1);
      }
      // Complex eigenvalue pair
      else if (j<nsave-1) {
	Complex LambdaA = Lambda(j);
	Complex LambdaB = Lambda(j+1);
	if (LambdaA != conj(LambdaB)) {
	  cerr << "Warning! Non-conjugate complex eigenvalues!" << endl;
	  cerr << "abs, arg Lambda(" << sj1 << ") == "
	       << setw(prec+5) << abs(LambdaA) << ' '
	       << setw(prec+5) << arg(LambdaA) << endl;
	  cerr << "abs, arg Lambda(" << sj2 << ") == "
	       << setw(prec+5) << abs(LambdaB)
	       << setw(prec+5) << arg(LambdaB) << endl;
	}
	checkConjugacy(Vn.column(j), Vn.column(j+1));

	ColumnVector evA = real(Vn.column(j));
	ColumnVector evB = imag(Vn.column(j));
	//evA += real(Vn.column(j+1));
	//evB -= imag(Vn.column(j+1));

	FlowField efA(u); // set ef to have same size as u
	FlowField efB(u); // set ef to have same size as u
	vector2field(evA, efA, basis);
	vector2field(evB, efB, basis);
	Real c = 1.0/sqrt(L2Norm2(efA) + L2Norm2(efB));
	efA *= c;
	efB *= c;
	efA.save(outdir + "ef" + sj1);
	efB.save(outdir + "ef" + sj2);
	save(LambdaA, outdir + "eW" + sj1);
	save(LambdaB, outdir + "eW" + sj2);

	save(log(LambdaA) / T, outdir + "ew" + sj1);
	save(log(LambdaB) / T, outdir + "ew" + sj2);

	++j;

      }
    }
  }
}

// f^T(u) = time-T DNS integration of u
void sf(const FlowField& u, Real T, const FieldSymmetry& sigma, FlowField& sfu, 
	Real Reynolds, DNSFlags& flags, TimeStep& dt, int& fcount) {

  if (T<0) {
    cerr << "sf: negative integration time T == " << T << endl;
    cerr << "returning sigma f(u,T) == sigma u" << endl;
    sfu = u;
    sfu *= sigma;
    return;
  }

  // Special case #1: no time steps
  if (T == 0) {
    cout << "sf: T==0, no integration, returning" << endl;
    sfu = u;
    sfu *= sigma;
    return;
  }

  //cout << "sf(u,T) : L2Norm(u) == " << L2Norm(u) << ",  T == " << T << flush;

  const Real nu = 1.0/Reynolds;
  FlowField p(u.Nx(), u.Ny(), u.Nz(), 1, u.Lx(), u.Lz(), u.a(), u.b());
  sfu = u;

  // Adjust dt for CFL if necessary
  DNS dns(sfu, nu, dt, flags);
  if (dt.variable()) {
    dns.advance(sfu, p, 1);
    dt.adjust(dns.CFL());
    dns.reset_dt(dt);
    sfu = u;
  }

  //cout << ", CFL == " << dns.CFL() << endl;
  //cout << "t == " << flush;

  //  t == current time in integration
  //  T == total integration time
  // dT == CFL-check interval
  // dt == DNS time-step
  //  N == T/dT, n == dT/dt;
  //  T == N dT, dT == n dt
  //  t == s dT (s is loop index)
  cout << "f(u,t) : t == " << flush;
  for (int s=1; s<=dt.N(); ++s) {
    dns.advance(sfu, p, dt.n());
    Real t = s*dt.dT();
    if (s % 10 == 0)
      cout << iround(t) << flush;
    else if (s % 2 == 0)
      cout << '.' << flush;
    if (dt.adjust(dns.CFL()))
      dns.reset_dt(dt);
  }
  sfu *= sigma;

  cout << "  CFL == " << dns.CFL() << endl;
  ++fcount;
}

void Dsf(const FlowField& u, const FlowField& sfu, const FlowField& du, FlowField& Dsf_du,
	Real T, const FieldSymmetry& sigma, Real Reynolds, DNSFlags& flags,
	TimeStep& dt, Real eps, bool centerdiff, int& fcount) {

  FlowField u_du(du);
  FlowField sfu_du(du);

  if (!centerdiff) {

    cout << "Computing f^T(u + du) for approximation L du == f^T(u + du) - f^T(u)" << endl;
    u_du  = du;
    u_du *= eps;
    u_du += u;

    sfu_du = u_du;
    sf(u_du, T, sigma, sfu_du, Reynolds, flags, dt, fcount);
    //fu_du *= sigma;

    Dsf_du = sfu_du;
    Dsf_du -= sfu;

    cout << "L2Norm(u_du)             == " << L2Norm(u_du) << endl;
    cout << "L2Norm(du)               == " << L2Norm(du) << endl;
    cout << "L2Norm(sf(u))            == " << L2Norm(sfu) << endl;
    cout << "L2Norm(sf(u + du))       == " << L2Norm(sfu_du) << endl;
    cout << "L2Norm(sf(u + du)-sf(u)) == " << L2Norm(Dsf_du) << endl;

    Dsf_du *= 1.0/eps;
  }
  else {

    cout << "Computing f^T(u +/- du/2) for approximation L du == f^T(u + du/2) - f^T(u - du/2)" << endl;

    // Compute f^T(u+du)
    u_du  = du;
    u_du *= eps/2;
    u_du += u;

    sfu_du = u_du;
    sf(u_du, T, sigma, sfu_du, Reynolds, flags, dt, fcount);
    //sfu_du *= sigma;
    Dsf_du = sfu_du;

    // Compute f^T(u-du)
    u_du  = du;
    u_du *= -eps/2;
    u_du += u;

    sfu_du = u_du;
    sf(u_du, T, sigma, sfu_du, Reynolds, flags, dt, fcount);
    //sfu_du *= sigma;
    Dsf_du -= sfu_du;

    // Complete approximation L du == sigma f^T(u + du/2) - sigma f^T(u - du/2)
    Dsf_du *= 1.0/eps;
  }
}


void checkConjugacy(const ComplexColumnVector& u, const ComplexColumnVector& v) {
  ColumnVector f = real(u);
  ColumnVector g = real(v);
  Real rerr = L2Dist(f,g);
  f = imag(u);
  g = imag(v);
  g *= -1;
  Real ierr = L2Dist(f,g);
  if (rerr + ierr > 1e-14*(L2Norm(u)+L2Norm(v))) {
    cout << "error : nonconjugate u,v, saving these vectors" << endl;
    save(u, "uerr");
    save(v, "verr");
  }
}
