#include <sys/stat.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include <octave/oct.h>
#include <octave/CColVector.h>
#include <octave/dColVector.h>
#include <octave/dMatrix.h>
#include <octave/EIG.h>
#include <octave/dbleSVD.h>
#include "channelflow/flowfield.h"
#include "channelflow/diffops.h"
#include "channelflow/utilfuncs.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/periodicfunc.h"
#include "octutils.h"

// John Gibson, Wed Oct 10 16:10:46 EDT 2007

using namespace std;
using namespace channelflow;

typedef FieldSymmetry Symm;
template <class T> void push(const T& t, array<T>& arrayt);

Real meandissipation(const FlowField& u, Real T, Real Reynolds, const DNSFlags& dnsflags,
		     const TimeStep& dt, SolutionType solntype);

Real observable(FlowField& u);

// Return distance with sign along unit vector along phi
Real spythag(Real Lx, Real Lxtarg, Real Lz, Real Lztarg, Real phi);

int main(int argc, char* argv[]) {


  string purpose("(ALPHA VERSION!) continue invariant solution of plane Couette flow in Reynolds number");
  ArgList args(argc, argv, purpose);

  DNSFlags dnsflags;
  GMRESHookstepFlags srchflags;

  const bool restart    = args.getflag("-r",    "--restart",       "start from three previously computed solutions");
  const bool eqb        = args.getflag("-eqb",  "--equilibrium",   "search for equilibrium or relative equilibrium (trav wave)");
  const bool orb        = args.getflag("-orb",  "--periodicorbit", "search for periodic orbit or relative periodic orbit");
  const bool poincsrch  = args.getflag("-poinc",  "--poincare",    "(relative) periodic orbit search constrained to I-D=0 Poincare section");
  srchflags.xrelative   = args.getflag("-xrel",  "--xrelative",    "search over x phase shift for relative orbit or eqb");
  srchflags.zrelative   = args.getflag("-zrel",  "--zrelative",    "search over z phase shift for relative orbit or eqb");

  const bool Recontinue = args.getflag("-contRe",    "--continueRe", "continue solution in Reynolds number");
  const bool Pcontinue  = args.getflag("-contdPdx",  "--continuedPdx", "continue solution in imposed pressure gradient");
  const bool Lxcontinue = args.getflag("-contLx",    "--continueLx", "continue solution in streamwise width Lx");
  const bool Lzcontinue = args.getflag("-contLz",    "--continueLz", "continue solution in spanwise width Lz");
  const bool aspectcont = args.getflag("-contAsp",   "--continueAspect",   "continue solution in aspect ratio Lx/Lz");
  const bool diagcont   = args.getflag("-contDiag",  "--continueDiagonal", "continue solution along diagonal with const aspect ratio Lx/Lz");
  const bool Ltargcont  = args.getflag("-contLtarg", "--continueLtarget",   "continue solution towards a target Lx,Lz");
  
  const bool upwards    = args.getflag("-up",  "--upwards",        "for non-restart searches, search in dir of increasing free parameter");
  /***/ Real Lxtarg     = args.getreal("-Lxtarg",  "--LxTarget",   2*pi, "aim for this value of Lx");
  /***/ Real Lztarg     = args.getreal("-Lztarg",  "--LzTarget",   pi, "aim for this value of Lz");
  const Real T0         = args.getreal("-T",  "--maptime",   15.0,  "initial guess for orbit period or time of eqb/reqb map f^T(u)");
  const Real Re0        = args.getreal("-R",  "--Reynolds", 400.0,  "Reynolds number");
  const Real dPdx0      = args.getreal("-dPdx", "--dPdx",     0.0,  "imposed mean pressure gradient");
  const  Real s0        = args.getreal("-s0", "--s0" ,        0.0,  "start value for arclength (arbitrary)");
  /****/ Real ds        = args.getreal("-ds", "--ds" ,        0.001, "initial arclength increment for quadratic extrapolation");

  /***/ bool adjdt      = args.getbool("-adt","--adjustDt",   true, "adjust dt between continuation steps to keep CFL roughly constant");

  const Real dsmin      = args.getreal("-dsmin", "--dsmin",   1e-04, "minimum arclength increment (in normalized D,Re space)");
  const Real dsmax      = args.getreal("-dsmax", "--dsmax",   0.05,  "maximum arclength increment (in normalized D,Re space)");
  const Real guesserrmin= args.getreal("-errmin", "--errmin", 1e-05, "minimum error for extrapolated guesses");
  const Real guesserrmax= args.getreal("-errmax", "--errmax", 1e-03, "maximum error for extrapolated guesses");

  const string sigmastr = args.getstr ("-sigma", "--sigma",      "", "file containing sigma of sigma f^T(u) - u = 0 (default == identity)");
  const string symmstr  = args.getstr ("-symms", "--symmetries", "", "file containing generators of isotropy group for symmetry-constrained search");

  dnsflags.baseflow     = PlaneCouette;
  dnsflags.timestepping = s2stepmethod(args.getstr("-ts", "--timestepping", "SBDF3", "time step algorithm: sbdf[1234]|cnab2|cnrk2|smrk2"));
  dnsflags.dealiasing   = DealiasXZ;
  dnsflags.nonlinearity = s2nonlmethod(args.getstr("-nl", "--nonlinearity", "Rotational", "nonlinearity method: rot|conv|skew|div|alt"));
  dnsflags.constraint   = PressureGradient;
  dnsflags.dPdx         = dPdx0;
  dnsflags.verbosity    = Silent;
  if (symmstr.length() > 0) {
    SymmetryList symms(symmstr);
    cout << "Restricting flow to invariant subspace generated by symmetries" << endl;
    cout << symms << endl;
    dnsflags.symmetries = symms;
  }


  const bool vardt     = args.getflag("-vdt","--variableDt", "vary dt during time integration to keep CFLmin<=CFL<CFLmax");
  //const bool autodt    = args.getflag("-sdt","--scaleDt",    "scale dt when Lx changes to keep dt/Lx const ");
  const Real dtarg     = args.getreal("-dt","--timestep",  0.03125, "(initial) integration timestep");
  const Real dtmin     = args.getreal("-dtmin", "--dtmin", 0.001, "minimum timestep");
  const Real dtmax     = args.getreal("-dtmax", "--dtmax", 0.04,  "maximum timestep");
  const Real dTCFL     = args.getreal("-dTCFL", "--dTCFL", 1.00,  "check CFL # at this interval");
  const Real CFLmin    = args.getreal("-CFLmin", "--CFLmin", 0.50, "minimum CFL number");
  const Real CFLmax    = args.getreal("-CFLmax", "--CFLmax", 0.60, "maximum CFL number");

  //const Real epsMax     = args.getreal("-emax","--epsMax",      1e-10, "minimum acceptable residual for GMRESHookstep");
  srchflags.epsSearch = args.getreal("-es",  "--epsSearch",   1e-10, "stop search if L2Norm(s f^T(u) - u) < epsSearch");
  srchflags.epsKrylov = args.getreal("-ek",  "--epsKrylov",   1e-13, "min. condition # of Krylov vectors");
  srchflags.epsDu     = args.getreal("-edu", "--epsDuLinear", 1e-5,  "relative size of du to u in linearization");
  srchflags.epsDt     = args.getreal("-edt", "--epsDtLinear", 1e-7,  "size of dT in linearization of f^T about T");
  srchflags.epsGMRES  = args.getreal("-eg",  "--epsGMRES",    1e-3,  "stop GMRES iteration when Ax=b residual is < this");
  srchflags.epsGMRESf = args.getreal("-egf",  "--epsGMRESfinal",  0.05,  "accept final GMRES iterate if residual is < this");
  srchflags.centdiff  = args.getflag("-cd", "--centerdiff", "centered differencing to estimate differentials");

  srchflags.Nnewton   = args.getint( "-Nn", "--Nnewton",   10,    "max number of Newton steps ");
  srchflags.Ngmres    = args.getint( "-Ng", "--Ngmres",    60,    "max number of GMRES iterations per restart");
  srchflags.Nhook     = args.getint( "-Nh", "--Nhook",     20,    "max number of hookstep iterations per Newton");

  srchflags.delta     = args.getreal("-d",    "--delta",      0.01,  "initial radius of trust region");
  srchflags.deltaMin  = args.getreal("-dmin", "--deltaMin",   0.009, "stop if radius of trust region gets this small");
  srchflags.deltaMax  = args.getreal("-dmax", "--deltaMax",   0.1,   "maximum radius of trust region");
  srchflags.deltaFuzz = args.getreal("-df",   "--deltaFuzz",  0.01,  "accept steps within (1+/-deltaFuzz)*delta");
  srchflags.lambdaMin = args.getreal("-lmin",   "--lambdaMin", 0.2,  "minimum delta shrink rate");
  srchflags.lambdaMax = args.getreal("-lmax",   "--lambdaMax", 1.5,  "maximum delta expansion rate");
  srchflags.lambdaRequiredReduction = 0.5; // when reducing delta, reduce by at least this factor.
  srchflags.improvReq = args.getreal("-irq",  "--improveReq", 1e-3,  "reduce delta and recompute hookstep if improvement "
				      " is worse than this fraction of what we'd expect from gradient");
  srchflags.improvOk  = args.getreal("-iok",  "--improveOk",   0.10, "accept step and keep same delta if improvement "
				      "is better than this fraction of quadratic model");
  srchflags.improvGood= args.getreal("-igd",  "--improveGood", 0.75, "accept step and increase delta if improvement "
				      "is better than this fraction of quadratic model");
  srchflags.improvAcc = args.getreal("-iac",  "--improveAcc",  0.10, "recompute hookstep with larger trust region if "
				      "improvement is within this fraction quadratic prediction.");

  srchflags.TscaleRel = args.getreal("-Tsc", "--Tscale",     20,    "scale time in hookstep: |T| = Ts*|u|/L2Norm(du/dt)");
  srchflags.dudtortho = args.getbool("-tc",  "--dudtConstr", true,  "require orthogonality of step to dudt");
  srchflags.orthoscale= args.getreal("-os",  "--orthoScale", 1,     "rescale orthogonality constraint");
  srchflags.dudxortho = args.getbool("-xc", "--dudxConstraint", true,"require orthogonality of step to dudx and dudz");
  srchflags.Rscale    = args.getreal("-rs", "--relativeScale", 1, "scale relative-search variables by this factor");
  srchflags.saveall   = args.getflag("-sa", "--saveall", "save each newton step");

  const string outdir   = args.getpath("-o",  "--outdir", "./", "output directory");
  const string logfile  = args.getstr ("-log","--logfile", "findsoln.log",  "output log (filename or \"stdout\")");
  const int digits      = args.getint("-dg", "--digits", 8, "number of digits for ReD.asc");

  string uname;
  string solndir[3];
  //Real ds0 = 0.01;

  const Real EPSILON = 1e-13;

  PoincareCondition* hpoincare = poincsrch ? new DragDissipation() : 0;
  TimeStep timestep(dtarg, dtmin, dtmax, dTCFL, CFLmin, CFLmax, vardt);

  if (!restart) {
    //ds0 = args.getreal("-ds0", "--ds0" ,  0.01, "arclength increment for perturbations of initial solution");
    //up  = args.getflag("-up", "--upwards",      "continue upwards in continuation param (Re, Lx, or Lz)");
    uname = args.getstr(1, "<flowfield>",       "initial guess for Newton search");
  }
  else {
    solndir[2]  = args.getpath(3, "<solndir0>", "directory containing previous solution data");
    solndir[1]  = args.getpath(2, "<solndir1>", "ditto");
    solndir[0]  = args.getpath(1, "<solndir2>", "ditto");
  }
  args.check();

  if (!(eqb^orb)) {
    cerr << "Please choose either -eqb or -orb option to search for (relative) equilibrium or (relative) periodic orbit" << endl;
    exit(1);
  }

  if (int(Recontinue) + int(Lxcontinue) + int(Lzcontinue) + int(aspectcont) + int(diagcont) + int(Pcontinue) + int(Ltargcont) != 1) {
    cerr << "Please choose one of -contRe, -contLx, -contLz -contAsp -contDiag, -contdPdx, or -contLxLzTarg options for continuation in Re, Lx, Lz, Lx/Lz, Lx*Lz, or dPdx, or target(Lx,Lz)" << endl;
    exit(1);
  }
  args.save("./");
  srchflags.solntype = orb ? PeriodicOrbit : Equilibrium;
  bool relative = srchflags.xrelative || srchflags.zrelative;
  bool Tnormalize = eqb;

  Symm sigma0;
  if (sigmastr.length() != 0)
    sigma0 = Symm(sigmastr);

  int W = 24;
  cout << setprecision(17);
  cout << "Working directory == " << pwd() << endl;
  cout << "Command-line args == ";
  for (int i=0; i<argc; ++i) cout << argv[i] << ' '; cout << endl;

  cout << "Re    == " << Re0 << endl;
  cout << "sigma == " << sigma0 << endl;
  cout << "T     == " << T0 << endl;
  cout << "dt    == " << timestep.dt() << endl;
  cout << "DNSFlags == " << dnsflags << endl << endl;

  array<FlowField> u(3); // solutions
  array<Real>      T(3); // period
  array<Real>      R(3); // Reynolds
  array<Real>      P(3); // pressure gradient, shorthand for dPdx
  array<Real>      D(3); // dissipation (avg for orbits)
  array<Real>     mu(3); // continuation parameter
  array<Real>    obs(3); // observable
  array<Real>    res(3); // residual
  array<Real>      s(3); // (mu,obs) arclengths
  array<Symm>  sigma(3); // symmetry
  
  Real CFL = 0.0; // CFL = c dt/dx;
  int fcount = 0;

  string muname = "mu";
  if (Recontinue)      muname = "mu=Re";
  else if (Lxcontinue) muname = "mu=Lx";
  else if (Lzcontinue) muname = "mu=Lz";
  else if (aspectcont) muname = "mu=Lx/Lz";
  else if (diagcont)   muname = "mu=sqrt(Lx^2+Lz^2)";
  else if (Pcontinue)  muname = "mu=dPdx";
  else if (Ltargcont)  muname = "mu=Ltarget";

  // Relations for aspect and diagonal continuation: 
  // Lx = diagonal * cos theta, Lz = diagonal * sin theta,
  Real aspect   = 0.0; // Lx/Lz
  Real diagonal = 0.0; // sqrt(Lx^2 + Lz^2)
  Real theta    = 0.0; // arccot(Lx/Lz) = arctan(Lz/Lx) = arctan(1/aspect);

  // Relations for target Lx,Lz continuation
  Real Ltarget = 0.0; // sqrt((Lxtarg-Lx0)^2 + (Lztarg-Lz0)^2), length to target
  Real phi = 0.0;     // arccot((Lxtarg-Lx0)/(Lztarg-Lz0)) 

  Real dt_Lx = 0.0;

  // Compute initial data points for extrapolation from perturbations of given solution
  if (!restart) {
    u[1] = FlowField(uname);
    u[2] = u[1];
    u[0] = u[1];

    R[2] = Re0;
    R[1] = Re0;
    R[0] = Re0;

    T[2] = T0;
    T[1] = T0;
    T[0] = T0;

    P[2] = dPdx0;
    P[1] = dPdx0;
    P[0] = dPdx0;


    aspect   = u[1].Lx()/u[1].Lz();
    diagonal = pythag(u[1].Lx(), u[1].Lz());
    theta    = atan(1.0/aspect);

    dt_Lx = dtarg/u[1].Lx();

    if (Ltargcont) {
      phi = atan((Lztarg-u[1].Lz())/(Lxtarg-u[1].Lx()));
      Ltarget = spythag(u[1].Lx(), Lxtarg, u[1].Lz(), Lztarg, phi);
    }
    else {
      Lxtarg = u[1].Lx();
      Lztarg = u[1].Lz();
      phi = 0.0;
      Ltarget  = 0.0;
    }

    cout << "Some geometrical parameters. Phi and Ltarg are used for continuation along line in 2d (Lx,Lz) plane." << endl;
    cout << "  phi == " << phi << endl;
    cout << "Ltarg == " << Ltarget << endl;
    cout << "distance  == " << pythag(Lxtarg-u[1].Lx(), Lztarg-u[1].Lz()) << endl;
    cout << "u[1].Lx() == " << u[1].Lx() << endl;
    cout << "   Lxtarg == " << Lxtarg << endl;
    cout << "u[1].Lz() == " << u[1].Lz() << endl;
    cout << "   Lztarg == " << Lztarg << endl;

    cout << "Lxtarg - Ltarg cos(phi) == " << Lxtarg - Ltarget*cos(phi) << endl;
    cout << "Lztarg - Ltarg sin(phi) == " << Lztarg - Ltarget*sin(phi) << endl;
    cout << "dt_Lx == " << dt_Lx << endl;

    const int musign = upwards ? 1 : -1; // Go up or down in continuation parameter?
    // Revise above as necessary
    if (Recontinue) {
      mu[2] = R[2] = Re0*(1-musign*ds);
      mu[1] = R[1] = Re0;
      mu[0] = R[0] = Re0*(1+musign*ds);
    }
    else if (Lxcontinue) {
      mu[2] = u[2].Lx() * (1-musign*ds);
      mu[1] = u[1].Lx();
      mu[0] = u[0].Lx() * (1+musign*ds);

      u[2].rescale(mu[2], u[2].Lz());
      u[0].rescale(mu[0], u[0].Lz());
    }
    else if (Lzcontinue) {
      mu[2] = u[2].Lz() * (1-musign*ds);
      mu[1] = u[1].Lz();
      mu[0] = u[0].Lz() * (1+musign*ds);

      u[2].rescale(u[2].Lx(), mu[2]);
      u[0].rescale(u[0].Lx(), mu[0]);
    }
    else if (aspectcont) {

      // Continuation parameter mu == Lx/Lz, keeping diagonal = sqrt(Lx^2 + Lz^2) fixed.
      // Let Lx' = diag cos theta'
      //     Lz' = diag sin theta'
      // Then diagonal is constant and aspect ratio is
      //     Lx'/Lz' = cot theta'
      //         mu' = cot theta'
      //      theta' = arccot mu' = arctan(1/mu') = arctan(Lz'/Lx')
      mu[2] = aspect*(1-musign*ds);
      mu[1] = aspect;
      mu[0] = aspect*(1+musign*ds);

      Real theta0 = atan(1/mu[0]);
      Real theta2 = atan(1/mu[2]);

      u[2].rescale(diagonal*cos(theta2), diagonal*sin(theta2));
      u[0].rescale(diagonal*cos(theta0), diagonal*sin(theta0));

    }
    else if (diagcont) {

      // Continuation parameter mu == sqrt(Lx^2 + Lz^2), keeping aspect ratio Lx/Lz constant
      // Let Lx' = k' Lx
      //     Lz' = k' Lz
      // Then aspect ratio is constant and
      //     mu' = sqrt(Lx'^2 + Lz'^2) = k' mu
      //      k' = mu'/mu;
      mu[2] = diagonal*(1-musign*ds);
      mu[1] = diagonal;
      mu[0] = diagonal*(1+musign*ds);

      u[2].rescale(mu[2]/mu[1]*u[1].Lx(), mu[2]/mu[1]*u[1].Lz());
      u[0].rescale(mu[0]/mu[1]*u[1].Lx(), mu[0]/mu[1]*u[1].Lz());
    }
    else if (Ltargcont) {

      // Continuation parameter mu = Ltarget = length to target Lx,Lz = pythag(Lxtarg-Lx, Lztarg-Lz)
      // Continuation is along l ine 
      // (Lx,Lz) = (Lxtarg - mu cos phi, Lztarg - mu sin phi)
      mu[2] = Ltarget*(1-musign*ds);
      mu[1] = Ltarget;
      mu[0] = Ltarget*(1+musign*ds);

      u[2].rescale(Lxtarg - mu[2]*cos(phi), Lztarg - mu[2]*sin(phi));
      u[1].rescale(Lxtarg - mu[1]*cos(phi), Lztarg - mu[1]*sin(phi));
      u[0].rescale(Lxtarg - mu[0]*cos(phi), Lztarg - mu[0]*sin(phi));
      
    }
    else if (Pcontinue) {
      if (dPdx0 != 0.0) {
	mu[2] = P[2] = dPdx0*(1-musign*ds);
	mu[1] = P[1] = dPdx0;
	mu[0] = P[0] = dPdx0*(1+musign*ds);
      }
      else {
	mu[2] = P[2] = -musign*ds*2/R[1];
	mu[1] = P[1] =  0.0;
	mu[0] = P[0] =  musign*ds*2/R[1];
      }
    }
    sigma[2] = sigma0;
    sigma[1] = sigma0;
    sigma[0] = sigma0;
  }

  // Read in some solutions from previous computations
  else {

    for (int i=0; i<3; ++i) {
      u[i] = FlowField(solndir[i] + "ubest");
      load(T[i], solndir[i] + "Tbest");
      load(R[i], solndir[i] + "Reynolds");
      load(P[i], solndir[i] + "dPdx");
      sigma[i] = Symm(solndir[i] + "sigmabest");
    }

    if (!relative && (sigma[0] != sigma[1] || sigma[1] != sigma[2]))
       cferror("continuesoln error : initial symmetries should be equal for non-relative searchs");
     sigma0 = sigma[1];

    aspect   = u[1].Lx()/u[1].Lz();
    diagonal = pythag(u[1].Lx(), u[1].Lz());
    theta    = atan(1.0/aspect);
    Ltarget = spythag(u[1].Lx(), Lxtarg, u[1].Lz(), Lztarg, phi);

    if (Recontinue) {
      mu[2] = R[2];
      mu[1] = R[1];
      mu[0] = R[0];
    }
    else if (Lxcontinue) {
      mu[2] = u[2].Lx();
      mu[1] = u[1].Lx();
      mu[0] = u[0].Lx();
    }
    else if (Lzcontinue) {
      mu[2] = u[2].Lz();
      mu[1] = u[1].Lz();
      mu[0] = u[0].Lz();
    }
    else if (aspectcont) {
      mu[2] = u[2].Lx()/u[2].Lz();
      mu[1] = u[1].Lx()/u[1].Lz();
      mu[0] = u[0].Lx()/u[0].Lz();

      // Check that diagonal of u[2]'s is same as u[0]. Rescale Lx,Lz keeping aspect ratio
      // constant if diagonals are different. Normally reloaded aspect-continued data will
      // have constant diagonals, but do check and fix in order to restart data with some
      // small fuzz in diagonal.
      for (int n=2; n>=0; n -=2) {
	if (abs(diagonal - pythag(u[n].Lx(), u[n].Lz())) >= EPSILON) {
	  cout << "Diagonal of u[" << n << "] != diagonal of u[1]. Rescaling Lx,Lz...." << endl;
	  Real theta_n = atan(1/mu[n]);
	  u[n].rescale(diagonal*cos(theta_n), diagonal*sin(theta_n));
	}
      }
    }
    else if (diagcont) {
      mu[2] = pythag(u[2].Lx(), u[2].Lz());
      mu[1] = pythag(u[1].Lx(), u[1].Lz());
      mu[0] = pythag(u[0].Lx(), u[0].Lz());

      // Check that aspect ratio of u[2]'s is same u[0]. Rescale Lx,Lz keeping diagonal
      // constant if aspect ratios differ. Normally reloaded diagonl-continued data will
      // have constant aspect ratio, but do check and fix in order to restart data with some
      // small fuzz in aspect ratio.
      for (int n=2; n>=0; n -=2) {
	if (abs(aspect - u[n].Lx()/u[n].Lz()) >= EPSILON) {
	  cout << "Aspect ratio of u[" << n << "] != aspect ratio of u[1]. Rescaling Lx,Lz...." << endl;
	  Real diagonal_n =  pythag(u[n].Lx(), u[n].Lz());
	  u[n].rescale(diagonal_n*cos(theta), diagonal_n*sin(theta));
	}
      }
    }
    else if (Ltargcont) {
      mu[2] = spythag(u[2].Lx(), Lxtarg, u[2].Lz(), Lztarg, phi);
      mu[1] = spythag(u[1].Lx(), Lxtarg, u[1].Lz(), Lztarg, phi);
      mu[0] = spythag(u[0].Lx(), Lxtarg, u[0].Lz(), Lztarg, phi);

      // Check that all solutions are colinear with (u[1].Lx, u[1].Lz) and (Lxtarg,Lztarg).
      // If not, rescale u[2] and u[0] Lx,Lz so that they are
      //for (int i=2; i>=0; i -=2) {
      //if (abs(atan((Lztarg - u[i].Lz())/(Lxtarg - u[i].Lx()))- phi) >= EPSILON) {
      //cout << "u[" << i << "] is not colinear with (u[1].Lx, u[1].Lz) and (Lxtarg,Lztarg). Rescaling to fix." << endl;
      //u[i].rescale(Lxtarg - mu[i]*cos(phi), Lztarg - mu[i]*sin(phi));
      //}
      //}
    }
    else if (Pcontinue) {
      mu[2] = P[2];
      mu[1] = P[1];
      mu[0] = P[0];
    }
    else {
      cerr << "Error in specification of continuation parameter! The control flow needs fixing (A)" << endl;
      exit(1);
    }

    dt_Lx = dtarg/u[1].Lx();

    cout << "loaded the following data..." << endl;
    cout << setw(4) << "i" << setw(W) << "T" << setw(W) << "Reynolds" << setw(W) << "dPdx" << setw(W) << "L2Norm(u)" << setw(W) << "sigma" << endl;
    for (int i=2; i>=0; --i)
      cout << setw(4) << i << setw(W) << T[i] << setw(W) << R[i] << setw(W) << P[i] << setw(W) << L2Norm(u[i]) << setw(W) << sigma[i] << endl;

  }

  if (aspectcont || diagcont) {
    cout << "aspect ratio || diagonal continuation : " << endl;
    cout << setprecision(15);
    cout.setf(ios::left);
    cout << setw(4) << "n" << setw(20) << muname << setw(20) << "aspect" << setw(20) << "diagonal" << endl;
    for (int n=2; n>=0; --n)
      cout << setw(4) << n << setw(20) << mu[n] << setw(20) << u[n].Lx()/u[n].Lz() << setw(20) << pythag(u[n].Lx(), u[n].Lz()) << endl;
    cout.unsetf(ios::left);
  }

  if (Ltargcont) {
    cout << "Lx,Lz target continuation : " << endl;
    cout << setprecision(15);
    cout.setf(ios::left);
    cout << setw(4) << "n" << setw(20) << muname << setw(20) << "Lx" << setw(20) << "Lz" << endl;
    for (int i=2; i>=0; --i)
      cout << setw(4) << i << setw(20) << mu[i] << setw(20) << u[i].Lx() << setw(20) << u[i].Lz() << endl;
    cout.unsetf(ios::left);
  }

  ofstream dos((outdir + "ReD.asc").c_str());
  dos << setw(W) << (string("%")+muname) << setw(W) << "D" << setw(W) << "s" << setw(W) << "guesserr" << setw(W) << "error" << " directory" << endl;
  dos << setprecision(digits);

  // Now find solutions for initial data.
  for (int i=2; i>=0; --i) {

    // Allow for more newton steps if we start with blind perturbation of Re.
    GMRESHookstepFlags srchflags0 = srchflags;
    if (!restart)
      srchflags0.Nnewton = 20;

    cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl;
    string searchdir = outdir + "restart-" + i2s(i) + "/";
    mkdir(searchdir);
    ofstream searchlog((searchdir + "findsoln.log").c_str());
    srchflags0.logstream = &searchlog;
    srchflags0.outdir    = searchdir;
    dnsflags.dPdx        = P[i];

    string label = "initial field u[" + i2s(i) + "]";
    fixdivnoslip(u[i]);
    project(dnsflags.symmetries, u[i], label, cout);

    cout << "computing solution for " << muname << " == " << mu[i] << " in dir " << searchdir << "..." << endl;
    cout << "dt    == " << timestep.dt() << endl;

    // Just use constant initial value of dt to start the continuation. Adjustment is for long term drift in CFL,dt
    res[i] = GMRESHookstep(u[i], T[i], sigma[i], R[i], hpoincare, srchflags0, dnsflags, timestep, CFL);
    D[i]   = meandissipation(u[i], T[i], R[i], dnsflags, timestep, srchflags0.solntype);
    obs[i] = D[i];

    cout << muname << " == " << mu[i] << "   obs == " << obs[i] << "   D == " << D[i] << "   searchresidual == " << res[i] << endl;
    save(R[i],  searchdir + "Reynolds");
    save(P[i],  searchdir + "dPdx");
    save(mu[i], searchdir + "mu");
    save(D[i],  searchdir + "D");
    sigma[i].save(searchdir + "sigmabest");
    save(CFL,  searchdir + "CFL");
    save(timestep.dt(),   searchdir + "dt");

    dos << setw(W) << mu[i] << setw(W) << D[i] << setw(W) << 0.0 << setw(W) << 0.0 << setw(W) << res[i] << " restart-"  << i << endl;
    cout << "CFL == " << CFL << endl;    
    cout << " dt == " << timestep.dt() << endl;
    if (adjdt) {
      timestep.adjustToMiddle(CFL);
      cout << "Readjusted dt" << endl;
      cout << " dt == " << timestep.dt() << endl;
    }    
  }

  Real obsnorm = abs(obs[1]);
  Real munorm  = abs(mu[1]);
  if (Pcontinue)
    munorm = 2/R[1]; // Poiseuille flow has dPdx = -2nu = -2/Reynolds, use this to normalize dPdx
  if (Ltargcont)
    munorm = pythag(Lxtarg,Lztarg); 

  cout << "     s0 == " << s0 << endl;
  cout << " munorm == " << munorm << endl;
  for (int i=2; i>=0; --i) 
    cout << " mu[" << i << "] == " << mu[i]  << endl;
  cout << "obsnorm == " << obsnorm << endl;
  for (int i=2; i>=0; --i) 
    cout << "obs[" << i << "] == " << obs[i]  << endl;

  s[2] = s0 - pythag((obs[2]-obs[1])/obsnorm, (mu[2]-mu[1])/munorm);
  s[1] = s0;
  s[0] = s0 + pythag((obs[0]-obs[1])/obsnorm, (mu[0]-mu[1])/munorm);

  //Real ds = sqrt(square((obs[0]-obs[1])/obsnorm) + square((mu[0]-mu[1])/munorm));

  const int Nsteps = 100;
  const Real guesserrtarget = sqrt(guesserrmin*guesserrmax); // aim between error bounds
  const int Ndsadjust = 6;

  bool prev_search_failed = false; // did previous search fail?

  Real guesserr_prev_search = guesserrtarget;
  Real guesserr = 0;

  for (int n=0; n<Nsteps; ++n) {

    cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl;
    cout << "Continuing previous solutions by pseudo-arclength quadratic extrapolation in s = (mu,obs)" << endl;
    cout << setw(8) << "i" << setw(W) << "s" << setw(W) << muname << setw(W) << "T" << setw(W) << "dPdx" << setw(W) << "obs" << endl;
    for (int i=2; i>=0; --i)
      cout << setw(8) << i << setw(W) << s[i] << setw(W) << mu[i] << setw(W) << T[i] << setw(W) << P[i] << setw(W) << obs[i] << endl;

    bool ds_reached_bounds = false;
    bool no_more_adjustments = false;

    Real munew;
    Real snew;
    Real Tnew;
    Real Rnew;
    Real Pnew;
    Symm sigmanew;
    FlowField unew;

    // Find a guess with error within error bounds, or under it if previous Newton search has failed
    for (int m=0; m<Ndsadjust; ++m) {

      cout << "------------------------------------------------------" << endl;
      cout << "Look for decent initial guess, step == " << m << endl;

      snew = s[0] + ds;
      Tnew      = quadraticInterpolate(T, s, snew);
      Rnew      = quadraticInterpolate(R, s, snew);
      Pnew      = quadraticInterpolate(P, s, snew);
      sigmanew  = quadraticInterpolate(sigma, s, snew);
      unew      = quadraticInterpolate(u, s, snew);

      // FlowField quadratic interpolation interpolates Lx and Lz independently.
      // Aspect and diagonal continuation and need to stay on a particular (Lx,Lz)
      // curve, so revise them.
      if (aspectcont) {
	Real theta_new = atan(unew.Lz()/unew.Lx());
	unew.rescale(diagonal*cos(theta_new), diagonal*sin(theta_new));
      }
      else if (diagcont) {
	Real diagonal_new = pythag(unew.Lx(), unew.Lz());
	unew.rescale(diagonal_new*cos(theta), diagonal_new*sin(theta));
      }
      fixdivnoslip(unew);
      //project(dnsflags.symmetries, unew, "quadratic extrapolant guess unew", cout);

      if (Recontinue)        munew = Rnew;
      else if (Lxcontinue)   munew = unew.Lx();
      else if (Lzcontinue)   munew = unew.Lz();
      else if (aspectcont)   munew = unew.Lx()/unew.Lz();
      else if (diagcont)     munew = pythag(unew.Lx(), unew.Lz());
      else if (Ltargcont)    munew = spythag(unew.Lx(), Lxtarg, unew.Lz(), Lztarg, phi);
      else if (Pcontinue)   {munew = Pnew; dnsflags.dPdx = Pnew;}
      else {
	cerr << "Error in specification of continuation parameter! The control flow needs fixing (B)" << endl;
	exit(1);
      }

      cout << setw(8) << "guess" << setw(W) << snew << setw(W) << munew << setw(W) << Tnew << setw(W) << Pnew << endl;

      cout << "calculating error of extrapolated guess" << flush;
      FlowField Guguess;
      G(unew, Tnew, sigmanew, Guguess, Rnew, dnsflags, timestep, Tnormalize, fcount, CFL);

      guesserr = L2Norm(Guguess);

      cout << endl;
      cout << "dt    == " << timestep.dt() << endl;
      cout << "CFL   == " << CFL << endl;
      cout << "dsmin == " << dsmin << endl;
      cout << "ds    == " << ds << endl;
      cout << "dsmax == " << dsmax << endl;
      cout << "guesserrmin == " << guesserrmin << endl;
      cout << "guesserr    == " << guesserr << endl;
      cout << "guesserrmax == " << guesserrmax << endl;
      cout << "guesserr previous  == " << guesserr_prev_search << endl;
      cout << "prev_search_failed == " << prev_search_failed << endl;
      cout << "ds_reached_bounds  == " << ds_reached_bounds << endl;
      cout << "no_more_adjustment == " << no_more_adjustments << endl;

      if (adjdt) {
	timestep.adjustToMiddle(CFL);
	cout << "Readjusted dt" << endl;
	cout << " dt == " << timestep.dt() << endl;
      }          

      // Decide whether to continue adjusting the guess, search on the guess, or give up.
      if (prev_search_failed && ds < dsmin) {
	cout << "Stopping continuation because continuing would require ds <= dsmin" << endl;
	exit(1);
      }
      else if (prev_search_failed && ds >= dsmin) {
	cout << "------------------------------------------------------" << endl;
	cout << "Stopping guess adjustments because previous search failed," << endl;
	cout << "so we're allowing guesserr <= guesserrmin, as long as dsmin <= ds" << endl;
	break;
      }
      // Previous search succeeded.
      if ((guesserr >= guesserrmin && guesserr <= guesserrmax)) {
	cout << "Stopping guess adjustments because guess error is in bounds: errmin <= guesserr <= errmax" << endl;
	break;
      }
      else if (ds_reached_bounds) {
	cout << "Stopping guess adjustments because ds has reached its bounds:" << endl;
	break;
      }
      else if (no_more_adjustments) {
	cout << "Stopping guess adjustments because a recent search failed and we're being cautious." << endl;
	break;
      }
      else if (guesserr_prev_search <= guesserrmin) {
	cout << "Previous search succeeded with guesserr <= guesserrmin. " << endl;
	cout << "Let's try increasing ds, to lesser of 1.5*ds and value suggested by guesserr = O(ds^3)" << endl;
	cout << "but no more than max allowed value dsmax. And no more guess adjustments after this." << endl;
	ds = lesser(1.5*ds, pow(guesserrmin/guesserr, 0.33));
	if (ds > dsmax) {
	  ds = dsmax;
	  ds_reached_bounds = true;
	}
	else
	  no_more_adjustments = true;
	continue;
      }
      else {
	cout << "Guess not yet within bounds. Try new guess based on model guesserr = O(ds^3)." << endl;
	// Adjust ds on model err = k O(ds^3), ds = 1/k err^1/3
	ds *= pow(guesserrtarget/guesserr, 0.33);
	if (ds < dsmin) {
	  ds = dsmin;
	  ds_reached_bounds = true;
	}
	else if (ds > dsmax) {
	  ds = dsmax;
	  ds_reached_bounds = true;
	}
	continue;
      }
    }

    // Done looking for a good guess. Now search on the guess.

    string searchdir = outdir + "search-" + i2s(n) + "/";
    mkdir(searchdir);
    save(Rnew, searchdir + "Reynolds");
    save(Pnew, searchdir + "dPdx");
    save(snew, searchdir + "arclength");
    save(munew, searchdir + "mu");
    save(timestep.dt(),   searchdir + "dt");
    ofstream searchlog((searchdir + "findsoln.log").c_str());

    //dnsflags.verbosity  = Silent;
    srchflags.logstream = &searchlog;
    srchflags.outdir    = searchdir;

    cout << "Computing new solution in directory " << searchdir << "  ..." << endl;
    Real resnew = GMRESHookstep(unew, Tnew, sigmanew, Rnew, hpoincare, srchflags, dnsflags, timestep, CFL);
    Real Dnew = meandissipation(unew, Tnew, Rnew, dnsflags, timestep, srchflags.solntype);
    Real obsnew = Dnew;

    cout << "dt    == " << timestep.dt() << endl;
    cout << "CFL   == " << CFL << endl;
    save(CFL,  searchdir + "CFL");


    if (resnew < srchflags.epsSearch) {
      cout << "Found new solution." << endl;

      save(Dnew,  searchdir + "D");
      save(Tnew,  searchdir + "Tbest");
      sigmanew.save(searchdir + "sigmabest");
      unew.save(searchdir + "ubest"); // resave u after symmetrization


      ds = sqrt(square((obsnew-obs[0])/obsnorm) + square((munew-mu[0])/munorm));
      snew = s[0]+ds;

      push(unew, u);
      push(Tnew, T);
      push(Rnew, R);
      push(Pnew, P);
      push(Dnew, D);
      push(munew, mu);
      push(obsnew, obs);
      push(resnew, res);
      push(snew, s);
      push(sigmanew, sigma);

      if (adjdt) {
	timestep.adjustToMiddle(CFL);
	cout << "Readjusted dt" << endl;
	cout << " dt == " << timestep.dt() << endl;
      }
      cout << setw(8) << "solution" << setw(W) << snew << setw(W) << munew << setw(W) << Tnew << "   residual == " << resnew << endl;

      dos << setw(W) << munew << setw(W) << Dnew << setw(W) << snew << setw(W) << guesserr << setw(W) << resnew << "  " << searchdir << endl;

      prev_search_failed = false;
      guesserr_prev_search = guesserr;

      
    }
    else {
      cout << "Failed to find new solution. Halving ds and trying again" << endl;
      cout << setw(8) << "failure" << setw(W) << snew << setw(W) << munew << setw(W) << Tnew << "   residual == " << resnew << endl;
      ds *= 0.5;
      prev_search_failed = true;
      guesserr_prev_search = guesserr;

      string failuredir(outdir + "failures");
      mkdir(failuredir);
      rename(searchdir, failuredir + "/search-" + i2s(n));
    }
  }
}

template <class T> void push(const T& t, array<T>& a) {
  for (int n=a.length()-1; n>0; --n)
    a[n] = a[n-1];
  a[0] = t;
}


//void temporalRegularize(FlowField& unew, Real Tnew, const FieldSymmetry& sigmanew, Real Rnew,
//			const DNSFlags& dnsflags, SolutionType solntype) {
//}

Real meandissipation(const FlowField& uarg, Real T, Real Reynolds, const DNSFlags& dnsflags,
		     const TimeStep& dtarg, SolutionType solntype) {

  cout << "computing mean dissipation..." << flush;
  FlowField u(uarg);

  ChebyCoeff U(u.Ny(),u.a(),u.b(),Spectral);
  U[1] = 1;

  if (solntype == Equilibrium) {
    u += U;
    cout << "done" << endl;
    return dissipation(u);
  }
  else {

    //DNSFlags dnsflags = dnsflagsarg;
    //dnsflags.verbosity = PrintTicks;

    int N = iround(T);
    const Real dT = T/N;
    TimeStep dt(dtarg.dt(), dtarg.dtmin(), dtarg.dtmax(), dT, dtarg.CFLmin(), dtarg.CFLmax(), dtarg.variable());
    u.optimizeFFTW(FFTW_PATIENT);

    const int Nx = u.Nx();
    const int Ny = u.Ny();
    const int Nz = u.Nz();
    const Real Lx = u.Lx();
    const Real Lz = u.Lz();
    const Real a = u.a();
    const Real b = u.b();

    FlowField q(Nx,Ny,Nz,1,Lx,Lz,a,b);

    DNS dns(u, 1/Reynolds, dt, dnsflags, 0.0);

    int count = 0;
    Real D = 0.0;

    // Don't close the orbit so as not to count last point twice
    for (Real t=0; t<=T-dT/2; t += dT) {

      u += U;
      D += dissipation(u);
      u -= U;
      ++count;

      dns.advance(u, q, dt.n());

      if (dt.adjust(dns.CFL()))
	dns.reset_dt(dt);
    }
    cout << "done" << endl;
    return D/count;
  }
}

Real observable(FlowField& u) {
  ChebyCoeff U(u.Ny(),u.a(),u.b(),Spectral);
  U[1] = 1;
  u += U;
  Real obs = dissipation(u);
  u -= U;
  return obs;
}

// distance with a sign
Real spythag(Real Lx, Real Lxtarg, Real Lz, Real Lztarg, Real phi) {

  Real ex = cos(phi);
  Real ez = sin(phi);

  Real dLx = Lxtarg-Lx;
  Real dLz = Lztarg-Lz;
  Real dist = pythag(dLx, dLz);
  Real projection = ex*dLx + ez*dLz;
  int sign = 0;
  if (projection > 0)
    sign = 1;
  else if (projection < 0)
    sign = -1;

  return sign*dist;
};

/* continuesoln.cpp, channelflow-1.3, www.channelflow.org
 *
 * Copyright (C) 2001-2009  John F. Gibson
 *
 * gibson@cns.physics.gatech.edu
 * Center for Nonlinear Sciences, School of Physics
 * Georgia Institute of Technology
 * Atlanta, GA 30332-0430
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, U
 */
