#include <sys/stat.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include <octave/oct.h>
#include <octave/CColVector.h>
#include <octave/dColVector.h>
#include <octave/dMatrix.h>
#include <octave/EIG.h>
#include <octave/dbleSVD.h>
#include "channelflow/flowfield.h"
#include "channelflow/diffops.h"
#include "channelflow/utilfuncs.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/periodicfunc.h"
#include "octutils.h"

using namespace std;
using namespace channelflow;

typedef FieldSymmetry Symm;
typedef FlowField     Field;

template <class T> void push(const T& t, array<T>& arrayt);

int main(int argc, char* argv[]) {

  string purpose("interpolate invariant solution u = sigma f^T(u), given three successive values of u,sigma,T");
  ArgList args(argc, argv, purpose);

  const Real NaN = 0.0/0.0;
  //cout << "NaN == " << NaN << endl;
  //cout << "isnan(NaN) == " << isnan(NaN) << endl;

  Real muRe    = args.getreal("-Re",  "--Reynolds", NaN, "interpolate with Reynolds number, use this value");
  Real muD     = args.getreal("-D",   "--dissipation", NaN, "interpolate with dissipation rate, use this value");
  Real muLx    = args.getreal("-Lx",  "--Lx",       NaN, "interpolate with Lx");
  Real muLz    = args.getreal("-Lz",  "--Lz",       NaN, "interpolate with Lz");
  Real muasp   = args.getreal("-asp", "--aspect",   NaN, "interpolate with aspect ratio Lx/Lz");
  Real mudiag  = args.getreal("-diag","--diagonal", NaN, "interpolate with diagonal");
  Real mudPdx  = args.getreal("-dPdx","--dPdx",     NaN, "interpolate with dPdx");

  const string odir     = args.getpath("-o", "--outdir", "./", "output directory");
  string solndir[3];
  solndir[2]  = args.getpath(3, "<solndir0>", "directory containing u0, T0, sigma0, Reynolds0, dPdx0");
  solndir[1]  = args.getpath(2, "<solndir1>", "directory containing u1, T1, sigma1, Reynolds1, dPdx1");
  solndir[0]  = args.getpath(1, "<solndir2>", "directory containing u2, T2, sigma2, Reynolds2, dPdx2");

  const Real EPSILON = 1e-13;

  args.check();

  int Nmu = !isnan(muRe) + !isnan(muD) + !isnan(muLx) + !isnan(muLz) + !isnan(muasp) + !isnan(mudiag) + !isnan(mudPdx);
  if (Nmu != 1) {
    cerr << "Please specify one and only one of the -Re -Lx -Lz -asp -diag and -dPdx options." << endl;
    exit(1);
  }
  args.save("./");

  cout << "Re == " << muRe << endl; 

  array<Field>  u(3); // solutions
  array<Real>   T(3); // period
  array<Real>   R(3); // Reynolds
  array<Real>   P(3); // pressure gradient, shorthand for dPdx
  array<Real>   D(3); // dissipation (avg for orbits)
  array<Symm>   s(3); // symmetry
  array<Real>  mu(3); // continuation parameter
  Real munew = NaN;

  for (int i=0; i<3; ++i) {
    u[i] = FlowField(solndir[i] + "ubest");
    load(T[i], solndir[i] + "Tbest");
    load(R[i], solndir[i] + "Reynolds");
    load(D[i], solndir[i] + "D");
    load(P[i], solndir[i] + "dPdx");
    s[i] = Symm(solndir[i] + "sigmabest");
  }
  Real aspect   = u[1].Lx()/u[1].Lz();
  Real diagonal = pythag(u[1].Lx(), u[1].Lz());
  Real theta    = atan(1.0/aspect);

  if (!isnan(muRe)) {
    mu[2] = R[2];
    mu[1] = R[1];
    mu[0] = R[0];
    munew = muRe;
  }
  else if (!isnan(muD)) {
    mu[2] = D[2];
    mu[1] = D[1];
    mu[0] = D[0];
    munew = muD;
  }
  else if (!isnan(muLx)) {
    mu[2] = u[2].Lx();
    mu[1] = u[1].Lx();
    mu[0] = u[0].Lx();
    munew = muLx;
  }
  else if (!isnan(muLz)) {
    mu[2] = u[2].Lz();
    mu[1] = u[1].Lz();
    mu[0] = u[0].Lz();
    munew = muLz;
  }
  else if (!isnan(muasp)) {
    mu[2] = u[2].Lx()/u[2].Lz();
    mu[1] = u[1].Lx()/u[1].Lz();
    mu[0] = u[0].Lx()/u[0].Lz();

    // Check that diagonal of u[2]'s is same as u[0]. Rescale Lx,Lz keeping aspect ratio
    // constant if diagonals are different. Normally reloaded aspect-continued data will
    // have constant diagonals, but do check and fix in order to restart data with some
    // small fuzz in diagonal.
    for (int n=2; n>=0; n -=2) {
      if (abs(diagonal - pythag(u[n].Lx(), u[n].Lz())) >= EPSILON) {
	cout << "Diagonal of u[" << n << "] != diagonal of u[1]. Rescaling Lx,Lz...." << endl;
	Real theta_n = atan(1/mu[n]);
	u[n].rescale(diagonal*cos(theta_n), diagonal*sin(theta_n));
      }
    }
    munew = muasp;
  }
  else if (!isnan(mudiag)) {
    mu[2] = pythag(u[2].Lx(), u[2].Lz());
    mu[1] = pythag(u[1].Lx(), u[1].Lz());
    mu[0] = pythag(u[0].Lx(), u[0].Lz());

    // Check that aspect ratio of u[2]'s is same u[0]. Rescale Lx,Lz keeping diagonal
    // constant if aspect ratios differ. Normally reloaded diagonl-continued data will
    // have constant aspect ratio, but do check and fix in order to restart data with some
    // small fuzz in aspect ratio.
    for (int n=2; n>=0; n -=2) {
      if (abs(aspect - u[n].Lx()/u[n].Lz()) >= EPSILON) {
	cout << "Aspect ratio of u[" << n << "] != aspect ratio of u[1]. Rescaling Lx,Lz...." << endl;
	Real diagonal_n =  pythag(u[n].Lx(), u[n].Lz());
	u[n].rescale(diagonal_n*cos(theta), diagonal_n*sin(theta));
      }
    }
    munew = mudiag;
  }
  else if (!isnan(mudPdx)) {
    mu[2] = P[2];
    mu[1] = P[1];
    mu[0] = P[0];
    munew = mudPdx;
  }
  else {
    cerr << "Error in specification of continuation parameter! The control flow needs fixing (A)" << endl;
    exit(1);
  }

  int W = 10;
  cout << "Loaded data..." << endl;
  cout.setf(ios::left);
  cout << setw(W) << "Re"
       << setw(W) << "D"
       << setw(W) << "Lx"
       << setw(W) << "Lz"
       << setw(W) << "dPdx"
       << setw(W) << "L2Norm(u)"
       << setw(W) << "T"
       << setw(W) << "sigma" << endl;
  for (int i=2; i>=0; --i) {
    cout.setf(ios::left);
    cout << setw(W) << R[i]
	 << setw(W) << D[i]
	 << setw(W) << u[i].Lx()
	 << setw(W) << u[i].Lz()
	 << setw(W) << P[i]
	 << setw(W) << L2Norm(u[i])
	 << setw(W) << T[i]
	 << setw(W) << s[i] << endl;
  }
  Real Tnew  = quadraticInterpolate(T, mu, munew);
  Real Rnew  = quadraticInterpolate(R, mu, munew);
  Real Dnew  = quadraticInterpolate(D, mu, munew);
  Real Pnew  = quadraticInterpolate(P, mu, munew);
  Symm snew  = quadraticInterpolate(s, mu, munew);
  Field unew = quadraticInterpolate(u, mu, munew);

  if (!isnan(muRe))
    Rnew = muRe;
  else if (!isnan(muD))
    Dnew = muD;
  else if (!isnan(mudPdx))
    Pnew = mudPdx;

  // FlowField quadratic interpolation interpolates Lx and Lz independently.
  // Aspect and diagonal continuation and need to stay on a particular (Lx,Lz)
  // curve, so revise them.
  if (!isnan(muasp)) {
    Real theta_new = atan(unew.Lz()/unew.Lx());
    unew.rescale(diagonal*cos(theta_new), diagonal*sin(theta_new));
  }
  else if (!isnan(mudiag)) {
    Real diagonal_new = pythag(unew.Lx(), unew.Lz());
    unew.rescale(diagonal_new*cos(theta), diagonal_new*sin(theta));
  }
  fixdivnoslip(unew);

  mkdir(odir);
  save(Tnew, odir + "T");
  save(Dnew, odir + "Dguess");
  save(Rnew, odir + "Reynolds");
  save(Pnew, odir + "dPdx");
  snew.save(odir + "sigma");
  unew.save(odir + "uguess");

  cout << "interpolated solution..." << endl;
  cout.setf(ios::left);

  cout << setw(W) << Rnew
       << setw(W) << Dnew
       << setw(W) << unew.Lx()
       << setw(W) << unew.Lz()
       << setw(W) << Pnew
       << setw(W) << L2Norm(unew)
       << setw(W) << Tnew
       << setw(W) << snew
       << endl;
}
  //project(dnsflags.symmetries, unew, "quadratic extrapolant guess unew", cout);
