#include <sys/stat.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include "channelflow/flowfield.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/utilfuncs.h"
#include "octutils.h"


using namespace std;
using namespace channelflow;

int main(int argc, char* argv[]) {

  string purpose("Check spatial and temporal resolution of a flowfield or solution of Navier-Stokes");
  ArgList args(argc, argv, purpose);

  DNSFlags dnsflags;

  const bool eqb        = args.getflag("-eqb",  "--equilibrium",   "input field is (relative) equilibrium");
  const bool orb        = args.getflag("-orb",  "--periodicorbit", "input field is (realtive) periodic orbit");
  const bool poincare   = args.getflag("-poinc",  "--poincare",    "(relative) periodic orbit search constrained to I-D=0 Poincare section");
  /***/ Real T          = args.getreal("-T", "--maptime",  10.0,  "initial guess for orbit period or time of eqb/reqb map f^T(u)");
  const Real Reynolds   = args.getreal("-R", "--Reynolds", 400.0, "Reynolds number");
  const string sigstr   = args.getstr ("-sigma", "--sigma", "",    "file containing symmetry of relative solution (default == identity)");
  const string symmstr   = args.getstr ("-symms", "--symmetries", "", "file containing generators of isotropy group for symmetry-constrained search");

  const int Nx0         = args.getint("-Nx", "--Nx", 0, "new x gridsize (0 => no change)");
  const int Ny0         = args.getint("-Ny", "--Ny", 0, "new y gridsize (0 => no change)");
  const int Nz0         = args.getint("-Nz", "--Nx", 0, "new z gridsize (0 => no change)");
  const Real dtarg      = args.getreal("-dt","--timestep",  0.03125, "integration timestep");
  const bool fixdiv     = args.getflag("-fdv",  "--fixdiv",   "ensure zero divergence and BCs on interpolated fields");
  dnsflags.baseflow     = PlaneCouette;
  dnsflags.timestepping = s2stepmethod(args.getstr("-ts", "--timestepping", "SBDF3", "time step algorithm: sbdf[1234]|cnab2|cnrk2|smrk2"));
  dnsflags.dealiasing   = DealiasXZ;
  dnsflags.nonlinearity = s2nonlmethod(args.getstr("-nl", "--nonlinearity", "Rotational", "nonlinearity method: rot|conv|skew|div|alt"));
  dnsflags.constraint   = PressureGradient;
  dnsflags.dPdx         = 0.0;
  dnsflags.verbosity    = Silent;

  const bool vardt     = args.getflag("-vdt","--variableDt", "adjust dt to keep CFLmin<=CFL<CFLmax");
  const Real dtmin     = args.getreal("-dtmin", "--dtmin", 0.001, "minimum timestep");
  const Real dtmax     = args.getreal("-dtmax", "--dtmax", 0.04,  "maximum timestep");
  const Real dTCFL     = args.getreal("-dTCFL", "--dTCFL", 1.00,  "check CFL # at this interval");
  const Real CFLmin    = args.getreal("-CFLmin", "--CFLmin", 0.40, "minimum CFL number");
  const Real CFLmax    = args.getreal("-CFLmax", "--CFLmax", 0.60, "maximum CFL number");
  const string uname    = args.getstr(1, "<flowfield>", "input field");

  args.check();
  args.save("./");

  PoincareCondition* hpoincare = poincare ? new DragDissipation() : 0;

  TimeStep dt(dtarg, dtmin, dtmax, dTCFL, CFLmin, CFLmax, vardt);
  FieldSymmetry sigma;
  if (sigstr.length() != 0)
    sigma = FieldSymmetry(sigstr);

  if (symmstr.length() > 0) {
    SymmetryList symms(symmstr);
    cout << "Restricting flow to invariant subspace generated by symmetries" << endl;
    cout << symms << endl;
    dnsflags.symmetries = symms;
  }

  FlowField u(uname);
  u.makeSpectral();

  const int Kx = u.padded() ? u.kxmaxDealiased() : u.kxmax();
  const int Kz = u.padded() ? u.kzmaxDealiased() : u.kzmax();
  const int Ky = u.My() - 1;

  // How large are the highest order spectral coefficients, in each direction?
  cout << "Truncation error estimated by the largest magnitude of the highest-order modes in x,y,z: " << endl;
  cout << "Expansion goes up to (Kx,Ky,Kz) == (+/-" << Kx << ',' << Ky << ',' << Kz << ')' << endl;

  // Find largest Fourier coefficient over (kx,kz) == (+/- Kx, *)
  Real maxmag_kx = 0.0;
  int  max_kx  = 0;
  int  max_kz  = 0;
  for (int kx = -Kx; kx<=Kx; kx += 2*Kx) {
    int mx = u.mx(kx);
    for (int kz=0; kz<Kz; ++kz) {
      int mz = u.mz(kz);
      Real mag_kxkz = L2Norm(u.profile(mx,mz));
      if (mag_kxkz > maxmag_kx) {
	maxmag_kx = mag_kxkz;
	max_kx = kx;
	max_kz = kz;
      }
    }
  }
  cout << "x: L2Norm(u_{Kx,kz}(y)) == " << setw(15) << maxmag_kx << " for (Kx,kz) == (" << max_kx << ',' << max_kz << ')' << endl;

  // Find largest Fourier coefficient over (kx,kz) == (*, Kz)
  Real maxmag_kz = 0.0;
  int mz = u.mz(Kz);
  for (int kx = -Kx; kx<=Kx; ++kx) {
    int mx = u.mx(kx);
    Real mag_kxkz = L2Norm(u.profile(mx,mz));
    if (mag_kxkz > maxmag_kx) {
      maxmag_kz = mag_kxkz;
      max_kx = kx;
    }
  }
  cout << "z: L2Norm(u_{kx,Kz}(y)) == " << setw(15) << maxmag_kz << " for (kx,Kz) == (" << max_kx << ',' << Kz << ')' << endl;

  // Find largest ky=My Chebyshev coefficient over (kx,kz) = (*,*);
  Real maxmag_ky = 0.0;
  for (int i=0; i<u.Nd(); ++i) {
    for (int kx = -Kx; kx<=Kx; ++kx) {
      int mx = u.mx(kx);
      for (int kz=0; kz<=Kz; ++kz) {
	int mz = u.mz(Kz);
	Real mag_kxkz = abs(u.cmplx(mx,Ky,mz,i));
	if (mag_kxkz > maxmag_ky) {
	  maxmag_ky = mag_kxkz;
	  max_kx = kx;
	  max_kz = kz;
	}
      }
    }
  }
  cout << "y: |u_{kx,Ky,kz,i}|     == " << setw(15) << maxmag_ky << " for (kx,Ky,kz) == (" << max_kx << ',' << Ky << ',' << max_kz << ')' << endl;

  if (!orb && ! eqb)
    exit(0);


  const int Nx = (Nx0 == 0 ? u.Nx() : Nx0);
  const int Ny = (Ny0 == 0 ? u.Ny() : Ny0);
  const int Nz = (Nz0 == 0 ? u.Nz() : Nz0);

  FlowField u1(Nx,Ny,Nz, u.Nd(), u.Lx(), u.Lz(), u.a(),u.b());
  if (Nx != u.Nx() || Ny != u.Ny() || Nz != u.Nz()) {
    cout << "Changing gridsize Nx,Ny,Nz from "
	 << u.Nx() << ',' << u.Ny() << ',' << u.Nz() << " to "
	 << Nx << ',' << Ny << ',' << Nz << endl;
    u1.interpolate(u);
    u1.zeroPaddedModes();
    //if ((Ny < u0.Ny() || a != u0.a() || b != u0.b()) & fixdiv) {
    cout <<"L2Norm, divergence, and BCs of original and interpolated fields u, u1..." << endl;
    cout << "L2Norm(u)   == " << L2Norm(u) << endl;
    cout << "L2Norm(u1)  == " << L2Norm(u1) << endl;
    cout << "divNorm(u)  == " << divNorm(u) << endl;
    cout << "divNorm(u1) == " << divNorm(u1) << endl;
    cout << "bcNorm(u)   == " << bcNorm(u) << endl;
    cout << "bcNorm(u1)  == " << bcNorm(u1) << endl;

    if (fixdiv) {
      cout <<"fixing divergence and no-slip bcs..." << endl;
      Vector v;
      FlowField foo(Nx,Ny,Nz,u1.Nd(),u1.Lx(),u1.Lz(),u1.a(),u1.b());
      field2vector(u1,v);
      vector2field(v,foo);
      u1 = foo;
      cout << "L2Norm(u1)  == " << L2Norm(u1) << endl;
      cout << "divNorm(u1) == " << divNorm(u1) << endl;
      cout << "bcNorm(u1)  == " << bcNorm(u1) << endl;
    }
  }
  else
    u1 = u;


  FlowField Gu;
  int fcount = 0;
  Real CFL = 0.0;
  bool Tnormalize = eqb;
  cout << "Computing residual of solution on new grid" << endl;
  Gp(u1, T, Reynolds, hpoincare, sigma, Gu, dnsflags, dt, Tnormalize, fcount, CFL, cout);

  cout << (Tnormalize ? "\n1/T " : "\n") << "L2Norm(u - sigma f^T(u))  == " << L2Norm(Gu) << ", CFL == " << CFL << endl;
}
