#include <iostream>
#include <fstream>
#include <iomanip>
#include "channelflow/flowfield.h"
#include "channelflow/dns.h"
#include "channelflow/utilfuncs.h"

//#include "dnspoincare.h"

using namespace std;
using namespace channelflow;


int main(int argc, char* argv[]) {

  string purpose("integrate plane Couette or channel flow from a given initial condition until it hits\n"
		 "Poincare section defined by I-D == 0 (default) or (u(t)-ustar, estar) == 0 (-psect option)");

  ArgList args(argc, argv, purpose);

  const int crosssign  = args.getint("-cr", "--crosssign" ,  0, "req. sign of dh/dt<0 for crossing h(u)==0 (0 means either)");
  const string bdir    = args.getpath("-bd", "--bdir", "./", "dir of basis functions that define fundamental domain");
  const int    Nb      = args.getint("-Nb", "--Nbasis", 0, "number of basis fields");

  const bool planeSection = args.getflag("-psect", "--planeSection", "poincare section defined by (u(t) - ustar, estar) == 0");
  const string ustar_str  = args.getstr("-us", "--ustar", "ustar", "a point (field) on poincare section");
  const string estar_str  = args.getstr("-es", "--estar", "estar", "a normal (field), defines poincare section");

  const bool u0cross  = args.getflag("-u0c", "--u0cross", "count initial condition as 0th crossing");
  const Real T0       = args.getreal("-T0", "--T0", 0.0, "start time");
  const Real T1       = args.getreal("-T1", "--T1", 100, "end time");
  const bool vardt    = args.getflag("-vdt", "--variabledt", "adjust dt to keep CFLmin<=CFL<CFLmax");
  const Real dtarg    = args.getreal("-dt",     "--dt", 0.03125, "timestep");
  const Real dtmin    = args.getreal("-dtmin",  "--dtmin", 0.001, "minimum time step");
  const Real dtmax    = args.getreal("-dtmax",  "--dtmax", 0.05,  "maximum time step");
  const Real dT       = args.getreal("-dT",     "--dT", 1.0,      "Poincare crossing check interval");
  const Real CFLmin   = args.getreal("-CFLmin", "--CFLmin", 0.40, "minimum CFL number");
  const Real CFLmax   = args.getreal("-CFLmax", "--CFLmax", 0.60, "maximum CFL number");
  const Real minnorm  = args.getreal("-mn", "--minnorm", 0.01, "stop if chebyNorm(u) < minnorm");
  const string stepstr= args.getstr ("-ts", "--timestepping", "sbdf3", "timestepping algorithm, one of "
				     "[cnfe1 cnab2 smrk2 sbdf1 sbdf2 sbdf3 sbdf4]");
  const string nonlstr= args.getstr ("-nl", "--nonlinearity", "rot", "method of calculating nonlinearity, "
				     "one of [rot conv div skew alt]");
  const string symmstr= args.getstr ("-symms","--symmetries", "", "constrain u(t) to invariant symmetric subspace, argument is the filename for a file listing the generators of the isotropy group");

  const bool savesteps= args.getflag("-ss", "--savesteps", "save u(t) at intervals dT");
  const string outdir = args.getpath("-o", "--outdir", "data-poinc/", "output directory");
  const string label  = args.getstr ("-l", "--label", "u", "output field prefix");
  const Real Reynolds = args.getreal("-R", "--Reynolds", 400, "Reynolds number");
  const bool channel  = args.getflag("-c", "--channel", "channelflow instead of plane Couette");
  const bool bulkvel  = args.getflag("-b", "--bulkvelocity","hold bulk velocity fixed, not pressure gradient");
  const Real dPdx     = args.getreal("-P", "--dPdx",   0.0, "value for fixed pressure gradient");
  const Real Ubulk    = args.getreal("-U", "--Ubulk",  0.0, "value for fixed bulk velocity");

  const string uname  = args.getstr (1, "<flowfield>", "initial condition");


  args.check();
  args.save("./");
  mkdir(outdir);

  cout << "Constructing u and optimizing FFTW..." << endl;
  FlowField u(uname);
  //u.optimizeFFTW(FFTW_PATIENT);
  u.makeSpectral();
  FlowField q(u.Nx(), u.Ny(), u.Nz(), 1, u.Lx(), u.Lz(), u.a(), u.b());

  TimeStep dt(dtarg, dtmin, dtmax, dT, CFLmin, CFLmax, vardt);
  const bool inttime = (abs(dT - int(dT)) < 1e-12) && (abs(T0 - int(T0)) < 1e-12) ? true : false;
  const Real nu = 1.0/Reynolds;


  FlowField uU(u);
  ChebyCoeff U(u.Ny(), u.a(), u.b(), Spectral);
  U[1] = 1;
  uU += U;
  cout << "dissip (u)    == " << dissipation(u) << endl;
  cout << "forcing(u)    == " << forcing(u) << endl;
  cout << "energy (u)    == " << 0.5*L2Norm2(u) << endl;
  cout << "dissip (u+U)  == " << dissipation(uU) << endl;
  cout << "forcing(u+U)  == " << forcing(uU) << endl;
  cout << "energy (u+U)  == " << 0.5*L2Norm2(uU) << endl;  cout << "D(u) == " << dissipation(uU) << endl;
  // Construct time-stepping algorithm
  DNSFlags flags;
  flags.timestepping = s2stepmethod(stepstr);
  flags.dealiasing   = DealiasXZ;
  flags.nonlinearity = s2nonlmethod(nonlstr);
  flags.constraint   = bulkvel ? BulkVelocity : PressureGradient;
  flags.baseflow     = channel ? Parabolic : PlaneCouette;
  flags.dPdx         = dPdx;
  flags.Ubulk        = Ubulk;
  flags.verbosity    = Silent;

  if (symmstr.length() > 0) {
    SymmetryList symms(symmstr);
    cout << "Restricting flow to invariant subspace generated by symmetries" << endl;
    cout << symms << endl;
    flags.symmetries = symms;
  }

  // A basis set to define fundamental domain: (u(t), efund[n]) > 0
  // and symmetries that map crossings back into fundamental domain
  array<FlowField> efund(Nb);
  array<FieldSymmetry> sfund(Nb);
  for (int n=0; n<Nb; ++n) {
    efund[n] = FlowField(bdir + "e" + i2s(n));
    sfund[n] = FieldSymmetry(bdir + "s" + i2s(n));

    FlowField se = sfund[n](efund[n]);
    cout << "L2IP(e,se) == " << L2IP(efund[n], se) << " for n == " << n << endl;
  }

  PoincareCondition* h;
  if (planeSection) {
    FlowField estar(estar_str);
    FlowField ustar(ustar_str);
    h = new PlaneIntersection(ustar, estar);
  }
  else
    h = new DragDissipation();

  //cout << "h(u) == " << (*h)(u) << endl;

  DNSPoincare dns(u, efund, sfund, h, nu, dt, flags, T0);

  int pluscount = 0;
  int minuscount = 0;

  ofstream tmios((outdir+"tmi.asc").c_str());
  ofstream tplos((outdir+"tpl.asc").c_str());
  tmios << setprecision(16);
  tplos << setprecision(16);

  Real t = T0;

  if (u0cross) {

    if (savesteps)
      u.save(outdir + label + t2s(t, inttime));

    FlowField u0(u);
    Real hu0 = (*h)(u);
    dns.advance(u,q,dt.n());
    t += dt.dT();
    Real hu1 = (*h)(u);
    Real dhdt = (hu1-hu0)/dt.dT();

    cout << "Counting initial field as zeroth crossing" << endl;
    cout << "h(u0) == " << hu0 << endl;
    cout << "h(u1) == " << hu1 << endl;
    cout << "dhdt  == " << dhdt << endl;

    if (dhdt < 0) {
      string countstr = i2s(pluscount++);
      u0.save(outdir + "upl" + countstr);
      save(T0, outdir + "tpl" + countstr);
      tplos << countstr << '\t' << t << endl;
    }
    else {
      string countstr = i2s(minuscount++);
      u0.save(outdir + "umi" + countstr);
      save(T0, outdir + "tmi" + countstr);
      tmios << countstr << '\t' << t << endl;
    }
  }

  cout.setf(ios::left);
  cout << setw(10) << "t"
       << setw(12) << "h(u)"
       << setw(12) << "L2Norm(u)"
       << setw(12) << "CFL" << endl;

  for ( ; t<=T1; t += dT) {

    Real CFL = dns.CFL();
    int it = iround(t);
    Real hval = (*h)(u);
    Real nrm = L2Norm(u);

    cout.setf(ios::left);
    cout << setw(10) << it
	 << setw(12) << hval
	 << setw(12) << nrm
	 << setw(12) << CFL << endl;

    if (savesteps)
      u.save(outdir + label + t2s(t, inttime));

    bool crossed = dns.advanceToSection(u, q, dt.n(), crosssign);


    if (crossed) {
      FlowField ucross = dns.ucrossing();
      Real tcross = dns.tcrossing();
      //Real hcross = dns.hcrossing();
      int  scross = dns.scrossing();

      if (scross > 0) {
	string countstr = i2s(pluscount++);
	ucross.save(outdir + "upl" + countstr);
	save(tcross, outdir + "tpl" + countstr);
	tplos << countstr << '\t' << tcross << endl;
      }
      else {
	string countstr = i2s(minuscount++);
	ucross.save(outdir + "umi" + countstr);
	save(tcross, outdir + "tmi" + countstr);
	tmios << countstr << '\t' << tcross << endl;
      }
    }
    if (chebyNorm(u) < minnorm) {
      cerr << "Reached minimum chebyNorm(u). Exiting." << endl;
      break;
    }

    /**********************
    Real CFL = dns.CFL();
    int it = iround(t);
    if (CFL < CFLmin)
      cout << it << " < " << flush;
    else if (CFL > CFLmax)
      cout << it << " > " << flush;
    else
      cout << it << ' ' << flush;
    **********************/
  }
  cout << endl;
}
